@@ -3,6 +3,8 @@ from dexy.exceptions import InternalDexyProblem
 from dexy.exceptions import UserFeedback
 from dexy.utils import file_exists
 from dexy.utils import s
+import chardet
+import codecs
 import dexy.batch
 import dexy.doc
 import dexy.parser
@@ -79,7 +81,11 @@ class Wrapper(object):
         self.initialize_attribute_defaults()
         self.update_attributes_from_kwargs(kwargs)
         self.project_root = os.path.abspath(os.getcwd())
+        self.project_root_ts = "%s%s" % (self.project_root, os.sep)
         self.state = None
+        self.current_task = None
+        self.lookup_nodes = {} # map of shortcuts/keys to all nodes which can match
+        self.lookup_sections = {} # map of section names to nodes
         self.transition('new')
 
     def state_message(self):
@@ -198,7 +204,7 @@ class Wrapper(object):
         try:
             shutil.rmtree(self.trash_dir())
         except OSError as e:
-            if not "No such file or directory" in str(e):
+            if not "No such file or directory" in unicode(e):
                 raise
 
     def reset_work_cache_dir(self):
@@ -221,7 +227,6 @@ class Wrapper(object):
         try:
             for node in matches:
                 for task in node:
-                    self.current_task = task
                     task()
 
         except Exception as e:
@@ -230,16 +235,28 @@ class Wrapper(object):
             if self.debug:
                 raise
             else:
-                msg = "ERROR while running %s: %s\n" % (self.current_task.key, str(e))
+                if self.current_task:
+                    msg = u"ERROR while running %s: %s\n" % (self.current_task.key, unicode(e))
+                else:
+                    msg = u"ERROR: %s\n" % unicode(e)
                 sys.stderr.write(msg)
                 self.log.warn(msg)
 
         else:
-            self.transition('ran')
-            self.batch.end_time = time.time()
-            self.batch.save_to_file()
-            shutil.move(self.this_cache_dir(), self.last_cache_dir())
-            self.empty_trash()
+            self.after_successful_run()
+
+    def after_successful_run(self):
+        self.transition('ran')
+        self.batch.end_time = time.time()
+        self.batch.save_to_file()
+        shutil.move(self.this_cache_dir(), self.last_cache_dir())
+        self.empty_trash()
+        self.add_lookups()
+
+    def add_lookups(self):
+        for data in self.batch:
+            data.add_to_lookup_sections()
+            data.add_to_lookup_nodes()
 
     def bundle_docs(self):
         from dexy.node import BundleNode
@@ -579,6 +596,18 @@ class Wrapper(object):
         key = node.key_with_class()
         self.nodes[key] = node
 
+    def add_data_to_lookup_nodes(self, key, data):
+        if not key in self.lookup_nodes:
+            self.lookup_nodes[key] = []
+        if not data in self.lookup_nodes[key]:
+            self.lookup_nodes[key].append(data)
+
+    def add_data_to_lookup_sections(self, key, data):
+        if not key in self.lookup_sections:
+            self.lookup_sections[key] = []
+        if not data in self.lookup_sections[key]:
+            self.lookup_sections[key].append(data)
+
     def qualify_key(self, key):
         """
         A full node key is of the form alias:pattern where alias indicates
@@ -654,31 +683,34 @@ class Wrapper(object):
         Return an Abstract Syntax Tree with information about nodes to be
         processed.
         """
-        parser_aliases = sorted(dexy.parser.Parser.plugins.keys())
+        ast = dexy.parser.AbstractSyntaxTree(self)
+
+        processed_configs = []
+
+        for alias in self.parsers.split():
+            parser = dexy.parser.Parser.create_instance(alias, self, ast)
 
-        # collect all doc config files in project dir
-        config_files = []
-        for alias in parser_aliases:
             for filepath, fileinfo in self.filemap.iteritems():
                 if fileinfo['dir'] == '.' or self.recurse or self.is_explicit_config(filepath):
                     if os.path.split(filepath)[1] == alias:
                         self.log.info("using config file '%s'" % filepath)
-                        config_file_info = (fileinfo['ospath'], fileinfo['dir'], alias,)
-                        config_files.append(config_file_info)
 
-        # warn if we don't find any configs
-        if len(config_files) == 0:
+                        config_file = fileinfo['ospath']
+                        dirname = fileinfo['dir']
+
+                        with open(config_file, "r") as f:
+                            config_text = f.read()
+
+                        try:
+                            processed_configs.append(filepath)
+                            parser.parse(dirname, config_text)
+                        except UserFeedback:
+                            sys.stderr.write("Problem occurred while parsing %s\n" % config_file)
+                            raise
+
+        if len(processed_configs) == 0:
             msg = "didn't find any document config files (like %s)"
-            self.printmsg(msg % (", ".join(parser_aliases)))
-
-        # parse each config file and add to ast
-        ast = dexy.parser.AbstractSyntaxTree(self)
-
-        for config_file, dirname, alias in config_files:
-            with open(config_file, "r") as f:
-                config_text = f.read()
-            parser = dexy.parser.Parser.create_instance(alias, self, ast)
-            parser.parse(dirname, config_text)
+            self.printmsg(msg % self.parsers)
 
         return ast
 
@@ -700,4 +732,15 @@ class Wrapper(object):
                 reporter.run(self)
 
     def is_location_in_project_dir(self, filepath):
-        return self.project_root in os.path.abspath(filepath)
+        return self.writeanywhere or (self.project_root_ts in os.path.abspath(filepath))
+
+    def decode_encoded(self, text):
+        if self.encoding == 'chardet':
+            encoding = chardet.detect(text)['encoding']
+            if not encoding:
+                return codecs.decode(text, "utf-8")
+            else:
+                return codecs.decode(text, encoding)
+        else:
+            return codecs.decode(text, self.encoding)
+
